/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this
 * file except in compliance with the License. You may obtain
 * a copy of the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an
 * "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is Java RASP toolkit.
 *
 * The Initial Developer of the Original Code is Lenio. Portions
 * created by Lenio are Copyright (C) 2007 Danish National IT and 
 * Telecom Agency (http://www.itst.dk). All Rights Reserved.
 *
 * Contributor(s):
 *   Tommy Dejbjerg Pedersen, Lenio
 *   Hans Guldager Knudsen, Lenio
 *   Patrik Johansson, Accenture
 *   Dennis S�gaard, Accenture
 *   Martin Bentzen, Accenture
 *   Jesper Jensen, Avanade
 *   Ramzi Fadel, Avanade
 *   Christian Pedersen, Accenture
 *   Morten Hougesen, Accenture
 *   Mikkel Hippe Brun, ITST
 *   Finn Hartmann Jordal, ITST
 *   Christian Lanng, ITST
 *
 */

package dk.firma.klient.webservice;

import java.io.FileInputStream;
import java.security.KeyStore;
import java.security.cert.X509Certificate;

import org.apache.axiom.om.OMElement;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import dk.gov.oiosi.common.Guid;
import dk.gov.oiosi.communication.Credentials;
import dk.gov.oiosi.communication.OiosiMessage;
import dk.gov.oiosi.communication.Request;
import dk.gov.oiosi.communication.Response;
import dk.gov.oiosi.communication.SendPolicy;
import dk.gov.oiosi.communication.Uri;
import dk.gov.oiosi.communication.client.TransportController;
import dk.gov.oiosi.communication.configuration.DocumentTypeConfig;
import dk.gov.oiosi.configuration.ConfigurationHandler;
import dk.gov.oiosi.raspprofile.communication.RaspRequest;
import dk.gov.oiosi.security.SignatureValidationProof;
import dk.gov.oiosi.security.oces.OcesX509Certificate;
import dk.gov.oiosi.xml.XmlDocument;
import dk.gov.oiosi.xml.XmlSchema;
import dk.gov.oiosi.xml.documenttype.DocumentTypeConfigSearcher;
import dk.gov.oiosi.xml.schema.SchemaValidator;
import dk.gov.oiosi.xml.schematron.SchematronValidator;

/**
 * Invoke RASP call against test RASP service defined by Service Skeleton.
 */
public class LocalTestClient {

    private static Log log = LogFactory.getLog(LocalTestClient.class);
    public static final String TEST_DOCUMENT_XML = "./resources/xml/OIOUBL_UtilityStatement_v2p2.xml";

    /**
     * Endpoint for testservice defined in Java/RASP Service skeleton.
     */
    private static final String ENDPOINT = "http://localhost:8080/dk.gov.oiosi.examples.service/services/TestService/nemhandel";

    /**
     * RASP Test application
     * 
     * This test client goes through the following steps:
     * 
     *  1. Load a document
     *  2. Validate the document
     *  3. Sends it to a service deployed on localhost (by default: the TestService service)
     */
    public static void main(String[] args) {
        try {
            log.info("Try to send message to service");

            // Set the path to the custom RASP configuration file. 
            // This file contains setup for things like:
            //   - Addresses to OCSP, LDAP and UDDI servers
            //   - The path to the Axis2 modules needed (RASP, WS-Security and WS-ReliableMessaging)
            //   - Configuration for the different OIO UBL document types
            ConfigurationHandler.setConfigurationFile("cfg/RaspConfiguration.xml");

            
            // Step 1: Load a document
            
            // Load the document from disc
            XmlDocument xmlDocument = new XmlDocument();
            xmlDocument.load(TEST_DOCUMENT_XML);
            log.info("Document created from file : ");

            // Search our configuration file for the document type we are currently about to send.
            // The UBL type can be inferred from the XML root element name of the file, and the configuration we get
            // contains paths to XML schemas, schematron stylesheets and XPath expressions used to find the EAN number for example
            DocumentTypeConfigSearcher documentTypeSearcher = new DocumentTypeConfigSearcher();
            DocumentTypeConfig documentTypeConfig = documentTypeSearcher.findUniqueDocumentType(xmlDocument);
            log.info("Document is known in configuration");

            // Step 2: Validate the document
            
            // Schema validate document before sending...
            SchemaValidator schemaValidator = new SchemaValidator();
            XmlSchema schema = new XmlSchema(documentTypeConfig.getSchemaPath());
            schemaValidator.schemaValidateXmlDocument(xmlDocument, schema);
            log.info("Document is schema valid");

            // Schematron validate Document before sending...
            SchematronValidator schematronValidator = new SchematronValidator(documentTypeConfig.getSchematronValidationConfig());
            schematronValidator.schematronValidateXmlDocument(xmlDocument);
            log.info("Document is schematron/logically valid");

            //Step 3: Send it
            
            Uri endpointAddress = new Uri(ENDPOINT);
            log.info("Destination service : " + endpointAddress.toString());

            // Load the client certificate from keystore
            KeyStore ks2 = KeyStore.getInstance("JKS");
            ks2.load(new FileInputStream("./keys/oiosi_test_certs.jks"), "Test1234".toCharArray());
            X509Certificate client = (X509Certificate) ks2.getCertificate("employee");
            
            // Load the server certificate from keystore
            KeyStore ks = KeyStore.getInstance("JKS");
            ks.load(new FileInputStream("./keys/oiosi_test_certs.jks"), "Test1234".toCharArray());
            X509Certificate server = (X509Certificate) ks.getCertificate("function");

            OcesX509Certificate ocesCertificateClient = new OcesX509Certificate(client);
            OcesX509Certificate ocesCertificateServer = new OcesX509Certificate(server);
            Credentials credentials = new Credentials(ocesCertificateClient, ocesCertificateServer);
            log.info("Destination service certificate : " + server.getSubjectX500Principal());

            // Use a sending policy defined in the configuration file
            SendPolicy sendPolicy = ConfigurationHandler.getInstance().getSendPolicy();
            log.info("Sendpolicy loaded from configuration");

            // Create the request
            Request request = new Request(endpointAddress, credentials, sendPolicy);
            
            //request.setUseRM(false);
            //request.setUseWSS(false);
            
            // Enable proxy
            // Use this code to send through a proxy (eg TcpMon) for logging purposes
            //request.setProxyHost("localhost");
            //request.setProxyPort(2222);
            
            // Wrap in RaspRequest to add custom headers and SOAP Actions.
            RaspRequest raspRequest = new RaspRequest(request);
            log.info("Request created - send...");

            // Create message wrapper
            OiosiMessage oiosiMessage = new OiosiMessage(xmlDocument);
            
            // Send request to service
            Response response = raspRequest.getResponse(oiosiMessage, Guid.newGuid().toString());
 
            OMElement body = response.getResponseMessage().getMessageXml().getOMDocument();
            SignatureValidationProof signatureProof = response.getResponseMessage().getSignatureProof();
            
            log.info("Received response from service: " + (body == null ? "empty message" : body));
            log.info("SignatureValidationProof: " + signatureProof);
            
        } catch (Exception e) {
            log.error("Failed to send message to service : " + e, e);
        }

        try {
            // Let RM Session finish...
            Thread.sleep(60000);
            TransportController.shutdown();
            log.info("Done...");
        } catch (Exception igore) {
        }
    }
}